@@ -35,6 +35,10 @@
 
 #include "raqm.h"
 
+#if FRIBIDI_MAJOR_VERSION >= 1
+#define USE_FRIBIDI_EX_API
+#endif
+
 /**
  * SECTION:raqm
  * @title: Raqm
@@ -59,83 +63,73 @@
  *     const char *fontfile;
  *     const char *text;
  *     const char *direction;
- *     int i;
+ *     const char *language;
  *     int ret = 1;
  *
- *     raqm_t *rq = NULL;
- *     raqm_glyph_t *glyphs = NULL;
- *     size_t count;
- *     raqm_direction_t dir;
- *
- *     FT_Library ft_library = NULL;
+ *     FT_Library library = NULL;
  *     FT_Face face = NULL;
- *     FT_Error ft_error;
  *
- *
- *     if (argc < 4)
+ *     if (argc < 5)
  *     {
- *         printf ("Usage:\n FONT_FILE TEXT DIRECTION\n");
+ *         printf ("Usage: %s FONT_FILE TEXT DIRECTION LANG\n", argv[0]);
  *         return 1;
  *     }
  *
  *     fontfile =  argv[1];
  *     text = argv[2];
  *     direction = argv[3];
+ *     language = argv[4];
  *
- *     ft_error = FT_Init_FreeType (&ft_library);
- *     if (ft_error)
- *         goto final;
- *     ft_error = FT_New_Face (ft_library, fontfile, 0, &face);
- *     if (ft_error)
- *         goto final;
- *     ft_error = FT_Set_Char_Size (face, face->units_per_EM, 0, 0, 0);
- *     if (ft_error)
- *         goto final;
- *
- *     dir = RAQM_DIRECTION_DEFAULT;
- *     if (strcmp (direction, "-rtl") == 0)
- *       dir = RAQM_DIRECTION_RTL;
- *     else if (strcmp (direction, "-ltr") == 0)
- *       dir = RAQM_DIRECTION_LTR;
- *
- *     rq = raqm_create ();
- *     if (rq == NULL)
- *         goto final;
- *
- *     if (!raqm_set_text_utf8 (rq, text, strlen (text)))
- *         goto final;
- *
- *     if (!raqm_set_freetype_face (rq, face))
- *         goto final;
- *
- *     if (!raqm_set_par_direction (rq, dir))
- *         goto final;
- *
- *     if (!raqm_layout (rq))
- *         goto final;
- *
- *     glyphs = raqm_get_glyphs (rq, &count);
- *     if (glyphs == NULL)
- *         goto final;
- *
- *     for (i = 0; i < count; i++)
+ *     if (FT_Init_FreeType (&library) == 0)
  *     {
- *         printf ("%d %d %d %d %d %d\n",
- *                glyphs[i].index,
- *                glyphs[i].x_offset,
- *                glyphs[i].y_offset,
- *                glyphs[i].x_advance,
- *                glyphs[i].y_advance,
- *                glyphs[i].cluster);
+ *       if (FT_New_Face (library, fontfile, 0, &face) == 0)
+ *       {
+ *         if (FT_Set_Char_Size (face, face->units_per_EM, 0, 0, 0) == 0)
+ *         {
+ *           raqm_t *rq = raqm_create ();
+ *           if (rq != NULL)
+ *           {
+ *             raqm_direction_t dir = RAQM_DIRECTION_DEFAULT;
+ *
+ *             if (strcmp (direction, "r") == 0)
+ *               dir = RAQM_DIRECTION_RTL;
+ *             else if (strcmp (direction, "l") == 0)
+ *               dir = RAQM_DIRECTION_LTR;
+ *
+ *             if (raqm_set_text_utf8 (rq, text, strlen (text)) &&
+ *                 raqm_set_freetype_face (rq, face) &&
+ *                 raqm_set_par_direction (rq, dir) &&
+ *                 raqm_set_language (rq, language, 0, strlen (text)) &&
+ *                 raqm_layout (rq))
+ *             {
+ *               size_t count, i;
+ *               raqm_glyph_t *glyphs = raqm_get_glyphs (rq, &count);
+ *
+ *               ret = !(glyphs != NULL || count == 0);
+ *
+ *               printf("glyph count: %zu\n", count);
+ *               for (i = 0; i < count; i++)
+ *               {
+ *                   printf ("gid#%d off: (%d, %d) adv: (%d, %d) idx: %d\n",
+ *                           glyphs[i].index,
+ *                           glyphs[i].x_offset,
+ *                           glyphs[i].y_offset,
+ *                           glyphs[i].x_advance,
+ *                           glyphs[i].y_advance,
+ *                           glyphs[i].cluster);
+ *               }
+ *             }
+ *
+ *             raqm_destroy (rq);
+ *           }
+ *         }
+ *
+ *         FT_Done_Face (face);
+ *       }
+ *
+ *       FT_Done_FreeType (library);
  *     }
  *
- *     ret = 0;
- *
- * final:
- *     raqm_destroy (rq);
- *     FT_Done_Face (face);
- *     FT_Done_FreeType (ft_library);
- *
  *     return ret;
  * }
  * ]|
@@ -397,12 +391,17 @@ raqm_set_text (raqm_t         *rq,
                const uint32_t *text,
                size_t          len)
 {
-  if (!rq || !text || !len)
+  if (!rq || !text)
     return false;
 
+  rq->text_len = len;
+
+  /* Empty string, don’t fail but do nothing */
+  if (!len)
+    return true;
+
   free (rq->text);
 
-  rq->text_len = len;
   rq->text = malloc (sizeof (uint32_t) * rq->text_len);
   if (!rq->text)
     return false;
@@ -434,30 +433,41 @@ raqm_set_text_utf8 (raqm_t         *rq,
                     const char     *text,
                     size_t          len)
 {
-#ifdef _MSC_VER
-  uint32_t *unicode = _alloca (sizeof (uint32_t) * len);
-#else
-  uint32_t unicode[len];
-#endif
+  uint32_t *unicode;
   size_t ulen;
+  bool ok;
 
-  if (!rq || !text || !len)
+  if (!rq || !text)
     return false;
 
+  /* Empty string, don’t fail but do nothing */
+  if (!len)
+  {
+    rq->text_len = len;
+    return true;
+  }
+
   RAQM_TEST ("Text is: %s\n", text);
 
   rq->flags |= RAQM_FLAG_UTF8;
 
-  rq->text_utf8 = malloc (sizeof (char) * strlen (text));
+  rq->text_utf8 = malloc (sizeof (char) * len);
   if (!rq->text_utf8)
     return false;
 
-  memcpy (rq->text_utf8, text, sizeof (char) * strlen (text));
+  unicode = malloc (sizeof (uint32_t) * len);
+  if (!unicode)
+    return false;
+
+  memcpy (rq->text_utf8, text, sizeof (char) * len);
 
   ulen = fribidi_charset_to_unicode (FRIBIDI_CHAR_SET_UTF8,
                                      text, len, unicode);
 
-  return raqm_set_text (rq, unicode, ulen);
+  ok = raqm_set_text (rq, unicode, ulen);
+
+  free (unicode);
+  return ok;
 }
 
 /**
@@ -535,9 +545,12 @@ raqm_set_language (raqm_t       *rq,
   hb_language_t language;
   size_t end = start + len;
 
-  if (!rq || !rq->text_len)
+  if (!rq)
     return false;
 
+  if (!rq->text_len)
+    return true;
+
   if (rq->flags & RAQM_FLAG_UTF8)
   {
     start = _raqm_u8_to_u32_index (rq, start);
@@ -636,9 +649,12 @@ _raqm_set_freetype_face (raqm_t *rq,
                          size_t  start,
                          size_t  end)
 {
-  if (!rq || !rq->text_len)
+  if (!rq)
     return false;
 
+  if (!rq->text_len)
+    return true;
+
   if (start >= rq->text_len || end > rq->text_len)
     return false;
 
@@ -707,9 +723,12 @@ raqm_set_freetype_face_range (raqm_t *rq,
 {
   size_t end = start + len;
 
-  if (!rq || !rq->text_len)
+  if (!rq)
     return false;
 
+  if (!rq->text_len)
+    return true;
+
   if (rq->flags & RAQM_FLAG_UTF8)
   {
     start = _raqm_u8_to_u32_index (rq, start);
@@ -769,7 +788,13 @@ _raqm_shape (raqm_t *rq);
 bool
 raqm_layout (raqm_t *rq)
 {
-  if (!rq || !rq->text_len || !rq->text_info)
+  if (!rq)
+    return false;
+
+  if (!rq->text_len)
+    return true;
+
+  if (!rq->text_info)
     return false;
 
   for (size_t i = 0; i < rq->text_len; i++)
@@ -1015,13 +1040,11 @@ static bool
 _raqm_itemize (raqm_t *rq)
 {
   FriBidiParType par_type = FRIBIDI_PAR_ON;
-#ifdef _MSC_VER
-  FriBidiCharType *types = _alloca (sizeof (FriBidiCharType) * rq->text_len);
-  FriBidiLevel *levels = _alloca (sizeof (FriBidiLevel) * rq->text_len);
-#else
-  FriBidiCharType types[rq->text_len];
-  FriBidiLevel levels[rq->text_len];
+  FriBidiCharType *types;
+#ifdef USE_FRIBIDI_EX_API
+  FriBidiBracketType *btypes;
 #endif
+  FriBidiLevel *levels;
   _raqm_bidi_run *runs = NULL;
   raqm_run_t *last;
   int max_level;
@@ -1047,6 +1070,21 @@ _raqm_itemize (raqm_t *rq)
   }
 #endif
 
+  types = calloc (rq->text_len, sizeof (FriBidiCharType));
+#ifdef USE_FRIBIDI_EX_API
+  btypes = calloc (rq->text_len, sizeof (FriBidiBracketType));
+#endif
+  levels = calloc (rq->text_len, sizeof (FriBidiLevel));
+  if (!types || !levels
+#ifdef USE_FRIBIDI_EX_API
+      || !btypes
+#endif
+      )
+  {
+    ok = false;
+    goto done;
+  }
+
   if (rq->base_dir == RAQM_DIRECTION_RTL)
     par_type = FRIBIDI_PAR_RTL;
   else if (rq->base_dir == RAQM_DIRECTION_LTR)
@@ -1063,8 +1101,15 @@ _raqm_itemize (raqm_t *rq)
   else
   {
     fribidi_get_bidi_types (rq->text, rq->text_len, types);
+#ifdef USE_FRIBIDI_EX_API
+    fribidi_get_bracket_types (rq->text, rq->text_len, types, btypes);
+    max_level = fribidi_get_par_embedding_levels_ex (types, btypes,
+                                                     rq->text_len, &par_type,
+                                                     levels);
+#else
     max_level = fribidi_get_par_embedding_levels (types, rq->text_len,
                                                   &par_type, levels);
+#endif
 
    if (par_type == FRIBIDI_PAR_LTR)
      rq->resolved_dir = RAQM_DIRECTION_LTR;
@@ -1072,16 +1117,25 @@ _raqm_itemize (raqm_t *rq)
      rq->resolved_dir = RAQM_DIRECTION_RTL;
   }
 
-  if (max_level < 0)
-    return false;
+  if (max_level == 0)
+  {
+    ok = false;
+    goto done;
+  }
 
   if (!_raqm_resolve_scripts (rq))
-    return false;
+  {
+    ok = false;
+    goto done;
+  }
 
   /* Get the number of bidi runs */
   runs = _raqm_reorder_runs (types, rq->text_len, par_type, levels, &run_count);
   if (!runs)
-    return false;
+  {
+    ok = false;
+    goto done;
+  }
 
 #ifdef RAQM_TESTING
   RAQM_TEST ("Number of runs before script itemization: %zu\n\n", run_count);
@@ -1100,7 +1154,10 @@ _raqm_itemize (raqm_t *rq)
   {
     raqm_run_t *run = calloc (1, sizeof (raqm_run_t));
     if (!run)
-      return false;
+    {
+      ok = false;
+      goto done;
+    }
 
     if (!rq->runs)
       rq->runs = run;
@@ -1122,7 +1179,10 @@ _raqm_itemize (raqm_t *rq)
         {
           raqm_run_t *newrun = calloc (1, sizeof (raqm_run_t));
           if (!newrun)
-            return false;
+          {
+            ok = false;
+            goto done;
+          }
           newrun->pos = runs[i].pos + j;
           newrun->len = 1;
           newrun->direction = _raqm_hb_dir (rq, runs[i].level);
@@ -1150,7 +1210,10 @@ _raqm_itemize (raqm_t *rq)
         {
           raqm_run_t *newrun = calloc (1, sizeof (raqm_run_t));
           if (!newrun)
-            return false;
+          {
+            ok = false;
+            goto done;
+          }
           newrun->pos = runs[i].pos + j;
           newrun->len = 1;
           newrun->direction = _raqm_hb_dir (rq, runs[i].level);
@@ -1187,7 +1250,13 @@ _raqm_itemize (raqm_t *rq)
   RAQM_TEST ("\n");
 #endif
 
+done:
   free (runs);
+  free (types);
+#ifdef USE_FRIBIDI_EX_API
+  free (btypes);
+#endif
+  free (levels);
 
   return ok;
 }
@@ -1447,6 +1516,7 @@ _raqm_shape (raqm_t *rq)
     hb_buffer_set_script (run->buffer, run->script);
     hb_buffer_set_language (run->buffer, rq->text_info[run->pos].lang);
     hb_buffer_set_direction (run->buffer, run->direction);
+    hb_buffer_set_flags (run->buffer, HB_BUFFER_FLAG_BOT | HB_BUFFER_FLAG_EOT);
     hb_shape_full (run->font, run->buffer, rq->features, rq->features_len,
                    NULL);
   }
@@ -1460,16 +1530,14 @@ _raqm_u32_to_u8_index (raqm_t   *rq,
                        uint32_t  index)
 {
   FriBidiStrIndex length;
-#ifdef _MSC_VER
-  char *output = _alloca ((sizeof (uint32_t) * index) + 1);
-#else
-  char output[(sizeof (uint32_t) * index) + 1];
-#endif
+  char *output = malloc ((sizeof (uint32_t) * index) + 1);
 
   length = fribidi_unicode_to_charset (FRIBIDI_CHAR_SET_UTF8,
                                        rq->text,
                                        index,
                                        output);
+
+  free (output);
   return length;
 }
 
@@ -1479,16 +1547,14 @@ _raqm_u8_to_u32_index (raqm_t   *rq,
                        uint32_t  index)
 {
   FriBidiStrIndex length;
-#ifdef _MSC_VER
-  uint32_t *output = _alloca (sizeof (uint32_t) * (index + 1));
-#else
-  uint32_t output[index + 1];
-#endif
+  uint32_t *output = malloc (sizeof (uint32_t) * (index + 1));
 
   length = fribidi_charset_to_unicode (FRIBIDI_CHAR_SET_UTF8,
                                        rq->text_utf8,
                                        index,
                                        output);
+
+  free (output);
   return length;
 }
 
@@ -1587,7 +1653,7 @@ raqm_index_to_position (raqm_t *rq,
 found:
   if (rq->flags & RAQM_FLAG_UTF8)
     *index = _raqm_u32_to_u8_index (rq, *index);
-  RAQM_TEST ("The position is %d at index %ld\n",*x ,*index);
+  RAQM_TEST ("The position is %d at index %zu\n",*x ,*index);
   return true;
 }
 
@@ -1671,7 +1737,7 @@ raqm_position_to_index (raqm_t *rq,
         }
         if (_raqm_allowed_grapheme_boundary (rq->text[*index],rq->text[*index + 1]))
         {
-          RAQM_TEST ("The start-index is %ld  at position %d \n", *index, x);
+          RAQM_TEST ("The start-index is %zu  at position %d \n", *index, x);
             return true;
         }
 
@@ -1685,7 +1751,7 @@ raqm_position_to_index (raqm_t *rq,
           }
           *index += 1;
         }
-        RAQM_TEST ("The start-index is %ld  at position %d \n", *index, x);
+        RAQM_TEST ("The start-index is %zu  at position %d \n", *index, x);
         return true;
       }
       else
@@ -1699,7 +1765,7 @@ raqm_position_to_index (raqm_t *rq,
   else
     *index = rq->text_len;
 
-  RAQM_TEST ("The start-index is %ld  at position %d \n", *index, x);
+  RAQM_TEST ("The start-index is %zu  at position %d \n", *index, x);
 
   return true;
 }
